



pip check
IF %ERRORLEVEL% NEQ 0 exit /B 1
set OPENBLAS_NUM_THREADS=1
IF %ERRORLEVEL% NEQ 0 exit /B 1
set OMP_NUM_THREADS=1
IF %ERRORLEVEL% NEQ 0 exit /B 1
setx MPLBACKEND "Agg"
IF %ERRORLEVEL% NEQ 0 exit /B 1
set "SKIMAGE_TEST_STRICT_WARNINGS=0" & pytest --verbose --pyargs skimage -k "not (_not_a_real_test or test_uint16 or test_apply_parallel_rgb or test_thresholding or test_mpl_imshow or test_cut_normalized or test_ncut_stable_subgraph or test_reproducibility or test_analytical_moments_calculation or test_rolling_ball or test_blur_effect or test_data.py or test_gray or test_shapes)"
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
